package BinomHeapTest;
import BinomHeap.*;
import java.util.*;

public class BinomHeapTest {
	static final int smallHeapSize = 10;
	static final int largeHeapSize = 10000;
	static BinomHeap h = null;
	static BinomHeap h1 = null;
	static BinomHeap h2 = null;
	static BinomHeap h3 = null;
	static MyBinomHeap g = null;
	static MyBinomHeap g1 = null;
	static MyBinomHeap g2 = null;
	static MyBinomHeap g3 = null;
	static Random rand = new Random();
	static boolean b;
	static String s = "(default)";
	static int m,n;
	static int errors = 0;
	static int cumErrors = 0;
	static final int maxErrorLimit = 20;

	public static void main(String[] args) {
		
		try {
			System.out.println("Binomial Heap Test for ");
			System.out.println(BinomHeap.names());
			System.out.println(new Date().toString());
		} catch (Exception e) {
			System.out.println("Exception thrown by BinomHeap.names()");
			System.out.println(e.toString());
		}
		
		/*
		Basic operations
		*/
		
		System.out.println();
		System.out.println("Basic test");
		
		//check makeHeap
		System.out.print("checking makeHeap...");
		n = rand.nextInt();
		try {
			h = BinomHeap.makeHeap(n);
			System.out.println("ok");
		} catch (Exception e) {
			System.out.println();
			System.out.println("Exception thrown by makeHeap(int)");
			System.out.println(e.toString());
			System.out.println("Cannot create heap, aborting test");
			System.exit(0);
		}
		
		//isEmpty() should return false
		System.out.print("checking isEmpty...");
		try {
			if (h.isEmpty()) {
				System.out.println();
				System.out.println("isEmpty() returned true on nonempty heap");
				logError();
			} else {
				System.out.println("ok");
			}
		} catch (Exception e) {
			System.out.println();
			System.out.println("Exception thrown by isEmpty()");
			System.out.println(e.toString());
			errors++;
		}
		
		//toString() should return "1"
		System.out.print("checking toString...");
		try {
			if (!h.toString().equals("1")) {
				System.out.println();
				System.out.println("toString() returned incorrect value");
//				System.out.println("<" + s + ">"); //fixed 10/7 dk
				//modified to give more useful info on error 10/8 dk 
				System.out.println("<" + h.toString() + "> should be <1>");
				logError();
			} else {
				System.out.println("ok");
			}
		} catch (Exception e) {
			System.out.println();
			System.out.println("Exception thrown by toString()");
			System.out.println(e.toString());
			errors++;
		}
		
		//check findMin
		System.out.print("checking findMin...");
		try {
			m = h.findMin();
			if (!(m == n)) {
				System.out.println();
				System.out.println("findMin() returned incorrect value");
				System.out.println(m);
				logError();
			} else {
				System.out.println("ok");
			}
		} catch (Exception e) {
			System.out.println();
			System.out.println("Exception thrown by findMin()");
			System.out.println(e.toString());
			errors++;
		}
		
		//check deleteMin
		System.out.print("checking deleteMin...");
		try {
			h.deleteMin();
			System.out.println("ok");
		} catch (Exception e) {
			System.out.println();
			System.out.println("Exception thrown by deleteMin()");
			System.out.println(e.toString());
			errors++;
		}
		
		//empty heap -- isEmpty() should return true
		System.out.print("checking isEmpty...");
		try {
			if (!h.isEmpty()) {
				System.out.println();
				System.out.println("isEmpty() returned false on empty heap");
				logError();
			} else {
				System.out.println("ok");
			}
		} catch (Exception e) {
			System.out.println();
			System.out.println("Exception thrown by isEmpty()");
			System.out.println(e.toString());
			errors++;
		}
		
		//toString() should return ""
		System.out.print("checking toString...");
		try {
			if (!h.toString().equals("")) {
				System.out.println();
				System.out.println("toString() returned incorrect value");
//				System.out.println("<" + s + ">"); //fixed 10/7 dk
				//modified to give more useful info on error 10/8 dk 
				System.out.println("<" + h.toString() + "> should be <>");
				logError();
			} else {
				System.out.println("ok");
			}
		} catch (Exception e) {
			System.out.println();
			System.out.println("Exception thrown by toString()");
			System.out.println(e.toString());
			errors++;
		}
		
		//findMin error handling
		System.out.print("checking findMin error handling...");
		try {
			m = h.findMin();
			System.out.println();
			System.out.println("findMin() returned a value on empty heap");
			System.out.println(m);
			logError();
		} catch (EmptyHeapException e) {
			System.out.println("ok");
		} catch (Exception e) {
			  System.out.println();
			  System.out.println("Wrong exception thrown by findMin()");
			  System.out.println(e.toString());
			  errors++;
		}
		
		//deleteMin error handling
		System.out.print("checking deleteMin error handling...");
		try {
			h.deleteMin();
			System.out.println();
			System.out.println("deleteMin() succeeded on empty heap");
			logError();
		} catch (EmptyHeapException e) {
			System.out.println("ok");
		} catch (Exception e) {
			  System.out.println();
			  System.out.println("Wrong exception thrown by deleteMin()");
			  System.out.println(e.toString());
			  errors++;
		}
		
		System.out.println("Basic test complete");
		System.out.println(errors + " errors");
		
		/*
		Small heap test
		*/
		
		System.out.println();
		System.out.println("Small heap test");
		errors = 0;
		//check insert
		System.out.print("inserting...");
		createHeaps(smallHeapSize);
		if (errors == 0) {
			System.out.println("ok");
		} else {
			System.out.println("insert failed with " + errors + " errors");
		}
		cumErrors = errors;
		errors = 0;		
		//check delete
		System.out.print("deleting...");
		dump(h,g);
		if (errors == 0) {
			System.out.println("ok");
		} else {
			System.out.println("delete failed with " + errors + " errors");
		}
		cumErrors += errors;		
		System.out.println("Small heap test complete");
		System.out.println(cumErrors + " errors");
		
		/*
		Large heap test
		*/
		
		System.out.println();
		System.out.println("Large heap test");
		errors = 0;
		//check insert
		System.out.print("inserting...");
		createHeaps(largeHeapSize);
		if (errors == 0) {
			System.out.println("ok");
		} else {
			System.out.println("insert failed with " + errors + " errors");
		}
		cumErrors = errors;
		errors = 0;
		//check delete
		System.out.print("deleting...");
		dump(h,g);
		if (errors == 0) {
			System.out.println("ok");
		} else {
			System.out.println("delete failed with " + errors + " errors");
		}
		cumErrors += errors;
		System.out.println("Large heap test complete");
		System.out.println(cumErrors + " errors");
		
		/*
		Merge test
		*/
		
		System.out.println();
		System.out.println("Merge test");
		errors = 0;
		
		System.out.print("creating heaps...");
		createHeaps(smallHeapSize);
		h1 = h;
		g1 = g;
		createHeaps(1);
		h2 = h;
		g2 = g;
		dump(h2,g2);
		createHeaps(1);
		h3 = h;
		g3 = g;
		dump(h3,g3);
		createHeaps(smallHeapSize);
		if (errors == 0) {
			System.out.println("ok");
		} else {
			System.out.println("unable to create heaps, aborting test");
			System.exit(0);
		}
		System.out.print("checking merge...");
		try {
			h.merge(h1);
		} catch (Exception e) {
			System.out.println();
			System.out.println("Exception thrown by merge");
			System.out.println(e.toString());
			errors++;
		}
		g.merge(g1);
		dump(h,g);
		System.out.println((errors == 0)?"ok":"failed");
		cumErrors = errors;
		errors = 0;
		System.out.print("checking merge from empty heap...");
		createHeaps(smallHeapSize);
		if (errors == 0) {
			try {
				h.merge(h2);
			} catch (Exception e) {
				System.out.println();
				System.out.println("Exception thrown by merge");
				System.out.println(e.toString());
				errors++;
			}
			g.merge(g2);
			dump(h,g);
			System.out.println((errors == 0)?"ok":"failed");
		} else {
			System.out.println("Cannot create heap, aborting test");
		}
		cumErrors += errors;
		errors = 0;
		System.out.print("checking merge to empty heap...");
		createHeaps(smallHeapSize);
		if (errors == 0) {
			try {
				h3.merge(h);
			} catch (Exception e) {
				System.out.println();
				System.out.println("Exception thrown by merge");
				System.out.println(e.toString());
				errors++;
			}
			g3.merge(g);
			dump(h3,g3);
		} else {
			System.out.println("Cannot create heap, aborting test");
		}
		System.out.println((errors == 0)?"ok":"failed");
		cumErrors += errors;
		System.out.println("Merge test complete");
		System.out.println(cumErrors + " errors");

		/*
		Pathology test
		*/
		
		System.out.println();
		System.out.println("Pathology test");
		errors = 0;
		
		System.out.print("creating test heaps...");
		createHeaps(smallHeapSize);
		h1 = h;
		g1 = g;
		createHeaps(smallHeapSize);
		h2 = h;
		g2 = g;
		createHeaps(smallHeapSize);
		if (errors != 0) {
			System.out.println("unable to create heaps, aborting test");
			System.exit(0);
		}
		try {
			h.merge(h1);
		} catch (Exception e) {
			System.out.println();
			System.out.println("Exception thrown by merge");
			System.out.println(e.toString());
			errors++;
		}
		g.merge(g1);
		if (errors == 0) {
			System.out.println("ok");
		} else {
			System.out.println("unable to merge heaps, aborting test");
			System.exit(0);
		}
		System.out.print("checking disposal of source heap on merge...");
		try {
			if (h1.isEmpty()) {
				System.out.println("ok");
			} else {
				System.out.println();
				System.out.println("Warning: source heap nonempty after merge");
				logError();
			}				
		} catch (Exception e) {
			System.out.println();
			System.out.println("Exception thrown by isEmpty()");
			System.out.println(e.toString());
			errors++;
		}
		cumErrors = errors;
		if (errors != 0) {
			errors = 0;
			//source heap not properly disposed of--try to force crash
			System.out.println("checking for pointer interference...");
			m = 3*smallHeapSize;
			try {
				//check for pointer interference between h1 and h
				while (m > 0 && !h1.isEmpty()) {
					h1.deleteMin();
					m--;
				} 			
			} catch (Exception e) {
				System.out.println();
				System.out.println("Exception thrown by isEmpty() or deleteMin()");
				System.out.println(e.toString());
				errors++;
			}
			try {
				if (m == 0 || !h1.isEmpty()) {
					System.out.println();
					System.out.println("Source heap corrupted");
					logError();
				}
			} catch (Exception e) {
				System.out.println();
				System.out.println("Exception thrown by isEmpty()");
				System.out.println(e.toString());
				System.out.println("Aborting test");
				System.exit(0);
			}
			//we emptied h1 legally--see if it affected h
			dump(h,g);
			if (errors == 0) {
				System.out.println("ok");
			} else {
				System.out.println("pointer interference from merged heap");
				errors++;
			}
		}
		cumErrors += errors;
		errors = 0;
		
		//check self-merge
		System.out.print("checking self-merge...");
		try {
			h2.merge(h2);
		} catch (Exception e) {
			System.out.println();
			System.out.println("Exception thrown by h.merge(h)");
			System.out.println(e.toString());
			//non necessarily an error
		}
		//clone g2
		try {
			g = MyBinomHeap.makeHeap(g2.findMin());
			g1 = MyBinomHeap.makeHeap(g2.findMin());
			g2.deleteMin();
			while (!g2.isEmpty()) {
				g.insert(g2.findMin());
				g1.insert(g2.findMin());
				g2.deleteMin();
			}
		} catch (Exception e) {
			throw new Error();
		}
		//find out how they handled it
		//should have either cloned and merged
		//or done nothing
		try {
			s = h2.toString();
		} catch (Exception e) {
			System.out.println();
			System.out.println("Exception thrown by toString()");
			System.out.println(e.toString());
			System.out.println("Aborting test");
			System.exit(0);			
		}
		if (!s.equals(g.toString())) { //if cloned and merged, we do too
			g.merge(g1);
		}
		try {
			if (s.equals(g.toString())) { //in either case check integrity
				dump(h2,g);
			} else {
				logError();
			}
		} catch (TooManyErrorsException f) {}
		if (errors == 0) {
			System.out.println("ok");
		} else {
			System.out.println("Heap corrupted by self-merge");
		}
		
		cumErrors += errors;			
		System.out.println("Pathology test complete");
		System.out.println(cumErrors + " errors");
		
		System.out.println();
		System.out.println("done");		
	}			
	
	//create 2 identical heaps of specified size
	//results returned in global variables h,g
	static void createHeaps(int size) {
		n = rand.nextInt();
		try {
			n = rand.nextInt();
			h = BinomHeap.makeHeap(n);
		} catch (Exception e) {
			System.out.println();
			System.out.println("Exception thrown by makeHeap(int)");
			System.out.println(e.toString());
			try {
				logError();
			} catch (TooManyErrorsException f) {
				return;
			}
		}
		g = MyBinomHeap.makeHeap(n);		
		for (int i = 0; i < size; i++) {
			try {
				n = rand.nextInt();
				h.insert(n);
			} catch (Exception e) {
				System.out.println();
				System.out.println("Exception thrown by insert(int)");
				System.out.println(e.toString());
				try {
					logError();
				} catch (TooManyErrorsException f) {
					return;
				}
			}
			g.insert(n);
			try {
				if (h.isEmpty()) {
					System.out.println();
					System.out.println("isEmpty() returned true on nonempty heap");
					try {
						logError();
					} catch (TooManyErrorsException f) {
						return;
					}
				}
				if (!h.toString().equals(g.toString())) {
					System.out.println();
					System.out.println("toString() returned incorrect value");
//					System.out.println("<" + s + ">"); //fixed 10/7 dk
					//modified to give more useful info on error 10/8 dk 
					System.out.println("<" + h.toString() + "> should be <" + g.toString() + ">");
					try {
						logError();
					} catch (TooManyErrorsException f) {
						return;
					}
				}
			} catch (Exception e) {
				System.out.println();
				System.out.println("Exception thrown by isEmpty() or toString()");
				System.out.println(e.toString());
				try {
					logError();
				} catch (TooManyErrorsException f) {
					return;
				}
			}			
		}
	}

	//empty out 2 nonempty heaps, checking that they are identical
	static void dump(BinomHeap h, MyBinomHeap g) {
		try {
			s = h.toString();
		} catch (Exception e) {
			System.out.println();
			System.out.println("Unexpected Exception thrown by toString()");
			System.out.println(e.toString());
			try {
				logError();
			} catch (TooManyErrorsException f) {
				return;
			}
		}
		if (!s.equals(g.toString())) {
			System.out.println();
			System.out.println("toString() returned incorrect value");
			//modified to give more useful info on error 10/8 dk 
			System.out.println("<" + s + "> should be <" + g.toString() + ">");
			try {
				logError();
			} catch (TooManyErrorsException f) {
				return;
			}
		}
		while (!g.isEmpty()) {
			try {
				if (h.isEmpty()) {
					System.out.println();
					System.out.println("isEmpty() returned true on nonempty heap");
					try {
						logError();
					} catch (TooManyErrorsException f) {
						return;
					}
				}
				if (h.findMin() != g.findMin()) {
					System.out.println();
					System.out.println("findMin() returned incorrect value");
					try {
						logError();
					} catch (TooManyErrorsException f) {
						return;
					}
				}
			} catch (EmptyHeapException e) {
				System.out.println();
				System.out.println("Premature EmptyHeapException thrown by findMin()");
				System.out.println(e.toString());
				try {
					logError();
				} catch (TooManyErrorsException f) {
					return;
				}
			} catch (Exception e) {
				  System.out.println();
				  System.out.println("Unexpected Exception thrown by findMin()");
				  System.out.println(e.toString());
				  try {
					  logError();
				  } catch (TooManyErrorsException f) {
					  return;
				  }
			}
			try {
				g.deleteMin();
			} catch (Exception e) {
				throw new Error();
			}
			try {
				h.deleteMin();
			} catch (EmptyHeapException e) {
				System.out.println();
				System.out.println("Premature EmptyHeapException thrown by deleteMin()");
				System.out.println(e.toString());
				try {
					logError();
				} catch (TooManyErrorsException f) {
					return;
				}
			} catch (Exception e) {
				  System.out.println();
				  System.out.println("Unexpected Exception thrown by deleteMin()");
				  System.out.println(e.toString());
				  try {
					  logError();
				  } catch (TooManyErrorsException f) {
					  return;
				  }
			}
			try {
				s = h.toString();
			} catch (Exception e) {
				System.out.println();
				System.out.println("Unexpected Exception thrown by toString()");
				System.out.println(e.toString());
				try {
					logError();
				} catch (TooManyErrorsException f) {
					return;
				}
			}
			//modified 10/8 dk
			//some people did not set largestRank to -1 on empty heap
//			if (!s.equals(g.toString())) {
			if (!s.equals(g.toString()) && !g.isEmpty()) {
				System.out.println();
				System.out.println("toString() returned incorrect value");
				//modified to give more useful info on error 10/8 dk 
				System.out.println("<" + s + "> should be <" + g.toString() + ">");
				try {
					logError();
				} catch (TooManyErrorsException f) {
					return;
				}
			}
		}		
		try {
			m = h.findMin();
			System.out.println();
			System.out.println("findMin() returned a value on empty heap");
			System.out.println(m);
			try {
				logError();
			} catch (TooManyErrorsException f) {
				return;
			}
		} catch (EmptyHeapException e) {
		} catch (Exception e) {
			  System.out.println();
			  System.out.println("Wrong exception thrown by findMin()");
			  System.out.println(e.toString());
			  try {
				  logError();
			  } catch (TooManyErrorsException f) {
				  return;
			  }
		}
		try {
			h.deleteMin();
			System.out.println();
			System.out.println("deleteMin() succeeded on empty heap");
			try {
				logError();
			} catch (TooManyErrorsException f) {
				return;
			}
		} catch (EmptyHeapException e) {
		} catch (Exception e) {
			  System.out.println();
			  System.out.println("Wrong exception thrown by deleteMin()");
			  System.out.println(e.toString());
			  try {
				  logError();
			  } catch (TooManyErrorsException f) {
				  return;
			  }
		}
	}
	
	static void logError() throws TooManyErrorsException {
		errors++;
		if (errors > maxErrorLimit) {
			System.out.println("Aborting after " + maxErrorLimit + " errors");
			throw new TooManyErrorsException();
		}
	}	
}
